"""Tests for client module."""
from unittest import TestCase
from unittest.mock import Mock, call, patch, sentinel

import grpc._channel
from google.protobuf.empty_pb2 import Empty
from testfixtures import LogCapture

from frgal.client import GrpcClient
from frgal.decoder import GrpcDecoder
from frgal.exceptions import ServiceNotFound, UnknownError

from .proto.example_pb2 import Reply


class TestGrpcClient(TestCase):
    """Test GrpcClient."""

    def setUp(self):
        self.grpc_module_mock = Mock(spec=('coffeeStub', ))

        self.log_handler = LogCapture()
        self.addCleanup(self.log_handler.uninstall)

    def test_connect_insecure(self):
        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel) as channel_mock:
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock])

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc)])

    def test_connect_secure(self):
        with patch('frgal.client.grpc.secure_channel', return_value=sentinel.channel) as channel_mock:
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock],
                                credentials=sentinel.credentials)

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])

    def test_get_stub(self):
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel):
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock])

        self.assertEqual(client.get_stub('coffee'), sentinel.coffee_stub)
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(sentinel.channel)])

    def test_get_stub_cache(self):
        # Test stubs are cached
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel):
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock])

        # Get stub once
        client.get_stub('coffee')
        # Get stub twice
        self.assertEqual(client.get_stub('coffee'), sentinel.coffee_stub)
        # Check it was created only once
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(sentinel.channel)])

    def test_get_stub_unknown(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock])

        with self.assertRaisesRegex(ServiceNotFound, 'Service teaStub could not be found'):
            client.get_stub('tea')

    def test_call(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = reply

        self.assertEqual(client.call('coffee', 'make', sentinel.request), 'Espresso!')

    def test_call_error(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        reply = Reply()
        reply.exception.tea_pot_error.CopyFrom(Reply.Exception.TeaPotError())
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = reply

        with self.assertRaisesRegex(UnknownError, 'tea_pot_error'):
            client.call('coffee', 'make', sentinel.request)

    def test_call_empty(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        reply = Empty()
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.return_value = reply

        self.assertIsNone(client.call('coffee', 'make', sentinel.request))

    def test_call_grpc_unavailable(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = grpc._channel._Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = [error, reply]

        with patch('frgal.client.grpc.insecure_channel') as channel_mock:
            self.assertEqual(client.call('coffee', 'make', sentinel.request), 'Espresso!')

        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc)])
        self.assertEqual(len(self.log_handler.records), 1)
        self.assertEqual(self.log_handler.records[0].getMessage(),
                         'Connection to gRPC service coffee failed. Trying to reconnect.')

    def test_call_grpc_other_error(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNKNOWN, "Unknown error")
        error = grpc._channel._Rendezvous(state, None, None, None)
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.side_effect = error

        with self.assertRaisesRegex(grpc._channel._Rendezvous, 'status = StatusCode.UNKNOWN'):
            client.call('coffee', 'make', sentinel.request)
